package nachos.threads;

import nachos.machine.*;

/**
 * A <i>communicator</i> allows threads to synchronously exchange 32-bit
 * messages. Multiple threads can be waiting to <i>speak</i>,
 * and multiple threads can be waiting to <i>listen</i>. But there should never
 * be a time when both a speaker and a listener are waiting, because the two
 * threads can be paired off at this point.
 */
public class Communicator {
    /**
     * Allocate a new communicator.
     */
    public Communicator() {
	lock = new Lock();
	canSend = new Condition(lock);
	canRecv = new Condition(lock);
	canLeave = new Condition(lock);
    }

    /**
     * Wait for a thread to listen through this communicator, and then transfer
     * <i>word</i> to the listener.
     *
     * <p>
     * Does not return until this thread is paired up with a listening thread.
     * Exactly one listener should receive <i>word</i>.
     *
     * @param	word	the integer to transfer.
     */
    public void speak(int word) {
	lock.acquire();

	while (full)
	    canSend.sleep();
	buffer = word;
	full = true;

	canRecv.wake();
	canLeave.sleep();

	lock.release();
    }

    /**
     * Wait for a thread to speak through this communicator, and then return
     * the <i>word</i> that thread passed to <tt>speak()</tt>.
     *
     * @return	the integer transferred.
     */    
    public int listen() {
	int result;

	lock.acquire();

	while (!full) {
	    canSend.wake();
	    canRecv.sleep();
	}

	result = buffer;
	full = false;
	canLeave.wake();
	canSend.wake();

	lock.release();
	return result;
    }
    
    Lock lock;
    int buffer;
    boolean full = false;
    Condition canSend, canRecv, canLeave;
}
